extern crate cc;
#[macro_use]
extern crate serde_derive;
extern crate serde;
extern crate serde_xml_rs;

use serde_xml_rs::deserialize;
use std::env;
use std::io::{self, Write};
use std::path::{Path, PathBuf};
use std::process::Command;
use std::time::SystemTime;

#[derive(Debug, Deserialize)]
struct RCC {
    qresource: Vec<QResource>,
}

#[derive(Debug, Deserialize)]
struct QResource {
    prefix: String,
    file: Vec<String>,
}

/// run a commmand and return the standard output if the command ran ok
/// otherwise print an error and exit this program
fn run(cmd: &str, command: &mut Command) -> Vec<u8> {
    match command.output() {
        Err(e) => {
            eprintln!(
                "Could not run {}. Make sure {} is in your path: {}",
                cmd,
                cmd,
                e
            )
        }
        Ok(output) => {
            io::stderr().write(&output.stderr).expect(
                "Could not write to stderr.",
            );
            if output.status.success() {
                return output.stdout;
            }
            match output.status.code() {
                None => eprintln!("Process {} terminated by signal.", cmd),
                Some(code) => eprintln!("{} exited with status code: {}", cmd, code),
            }
        }
    }
    ::std::process::exit(-1);
}

/// query a Qt environment variable via qmake
fn qmake_query(var: &str) -> String {
    String::from_utf8(run("qmake", Command::new("qmake").args(&["-query", var])))
        .expect("qmake output was not valid UTF-8")
}

fn get_extreme_mtime(
    paths: &[&str],
    f: fn(SystemTime, SystemTime) -> SystemTime,
) -> Option<SystemTime> {
    let mut oldest_mtime = None;
    for path in paths {
        if let Ok(mtime) = Path::new(path).metadata().and_then(|m| m.modified()) {
            if let Some(ym) = oldest_mtime {
                oldest_mtime = Some(f(mtime, ym));
            } else {
                oldest_mtime = Some(mtime);
            }
        } else {
            return None;
        }
    }
    oldest_mtime
}

/// return the oldest mtime for the paths
/// if one of the mtimes cannot be read, return None
fn get_oldest_mtime(paths: &[&str]) -> Option<SystemTime> {
    get_extreme_mtime(paths, |a, b| if a > b { b } else { a })
}

/// return the youngest/newest mtime for the paths
/// if one of the mtimes cannot be read, return None
fn get_youngest_mtime(paths: &[&str]) -> Option<SystemTime> {
    get_extreme_mtime(paths, |a, b| if a > b { a } else { b })
}

/// Run moc to generate C++ code from a Qt C++ header
fn moc(header: &str, output: &str) {
    run("moc", Command::new("moc").args(&["-o", output, header]));
}

/// Run rcc to generate C++ code from a Qt resource file
fn rcc(rcfile: &str, output: &str) {
    run("rcc", Command::new("rcc").args(&["-o", output, rcfile]));
}

/// return true if a command should run.
/// It returns true if all inputs are present and if the inputs are newer
/// than the output or if the outputs do not exist yet.
fn should_run(input: &[&str], output: &[&str]) -> bool {
    // get the youngest input time
    let input_time = match get_youngest_mtime(input) {
        Some(time) => time,
        _ => return false,
    };
    let output_time = get_oldest_mtime(output);
    output_time.is_none() || output_time.unwrap() <= input_time
}

/// Parse the qrc file, panic if it fails
fn read_qrc(qrc: &str) -> RCC {
    let bytes = ::std::fs::read(qrc).expect(&format!("Could not read {}", qrc));
    deserialize(&bytes[..]).expect(&format!("could not parse {}", qrc))
}

/// Get the list of files that are listed in the qrc file
fn qrc_to_input_list<'a>(qrc: &'a str, rcc: &'a RCC) -> Vec<&'a str> {
    let mut list = Vec::new();
    list.push(qrc);
    for qresource in &rcc.qresource {
        for file in &qresource.file {
            list.push(&file);
        }
    }
    list
}

fn main() {
    let out_dir: PathBuf = env::var("OUT_DIR").unwrap().into();
    let qt_library_path = qmake_query("QT_INSTALL_LIBS");
    let qt_include_path = qmake_query("QT_INSTALL_HEADERS");
    let mut moc_bindings = out_dir.clone();
    moc_bindings.push("moc_Bindings.cpp");
    let moc_bindings = moc_bindings.to_str().unwrap();
    if should_run(&["cpp/Bindings.h"], &[moc_bindings]) {
        moc("cpp/Bindings.h", moc_bindings);
    }
    let mut qml_cpp = out_dir.clone();
    qml_cpp.push("qml.cpp");
    let qml_cpp = qml_cpp.to_str().unwrap();
    let qrc = read_qrc("qml.qrc");
    let qrc_inputs = qrc_to_input_list("qml.qrc", &qrc);
    if should_run(&qrc_inputs, &[qml_cpp]) {
        rcc("qml.qrc", qml_cpp);
    }
    if should_run(
        &["cpp/main.cpp", "cpp/Bindings.cpp", moc_bindings, qml_cpp],
        &["libmailmodel.a"],
    )
    {
        cc::Build::new()
            .include(qt_include_path.trim())
            .cpp(true)
            .file("cpp/main.cpp")
            .file("cpp/Bindings.cpp")
            .file(moc_bindings)
            .file(qml_cpp)
            .compile("libmailmodel.a");
    }
    println!("cargo:rustc-link-search={}", qt_library_path.trim());
    println!("cargo:rustc-link-lib=Qt5Core");
    println!("cargo:rustc-link-lib=Qt5Network");
    println!("cargo:rustc-link-lib=Qt5Gui");
    println!("cargo:rustc-link-lib=Qt5Qml");
}
