use errors::Result;
use imap::client::Client;
use imap::Mailbox;
use imap_proto::parse_response;
use imap_proto::types::{AttributeValue, MessageSection, Response, SectionPath};
use mail_store::{MailStore, Name};
use mail_store_state::MailStoreState;
use message::{Message, Uid};
use message_threads::MessageThreader;
use native_tls::{TlsConnector, TlsStream};
use nom::IResult;
use std::fmt;
use std::io::{Read, Write};
use std::net::TcpStream;
use std::sync::Arc;

pub struct ConnectionOnly(pub Client<Box<Stream>>);

impl ConnectionOnly {
    pub fn examine(mut self, folder: &str) -> Result<ConnectionAndFolder> {
        let mailbox = self.0.examine(folder)?;
        Ok(ConnectionAndFolder {
            client: self.0,
            folder: folder.to_string(),
            mailbox,
        })
    }
}

pub struct ConnectionAndFolder {
    pub client: Client<Box<Stream>>,
    pub folder: String,
    pub mailbox: Mailbox,
}

impl ConnectionAndFolder {
    pub fn examine(mut self, folder: &str) -> Result<ConnectionAndFolder> {
        if self.folder != folder {
            self.mailbox = self.client.examine(folder)?;
            self.folder = folder.to_string();
        }
        Ok(self)
    }
}

impl ::std::fmt::Display for Uid {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        self.0.fmt(f)
    }
}

#[derive(Deserialize)]
pub struct IMAPConfig {
    domain: String,
    username: String,
    password: String,
    port: u16,
    ssl: bool,
}

pub enum Connection {
    None,
    ConnectionOnly(ConnectionOnly),
    ConnectionAndFolder(ConnectionAndFolder),
}

pub struct IMAPStore {
    config: IMAPConfig,
    connection: Connection,
}

impl IMAPStore {
    pub fn new(config: IMAPConfig) -> IMAPStore {
        IMAPStore {
            config,
            connection: Connection::None,
        }
    }
    fn only_connection(&mut self) -> Result<ConnectionOnly> {
        let mut c = Connection::None;
        ::std::mem::swap(&mut c, &mut self.connection);
        Ok(ConnectionOnly(match c {
            Connection::None => connect(&self.config)?,
            Connection::ConnectionOnly(c) => c.0,
            Connection::ConnectionAndFolder(c) => c.client,
        }))
    }
    fn examine(&mut self, folder: &str) -> Result<ConnectionAndFolder> {
        let mut c = Connection::None;
        ::std::mem::swap(&mut c, &mut self.connection);
        Ok(match c {
            Connection::ConnectionAndFolder(c) => c.examine(folder)?,
            _ => self.only_connection()?.examine(folder)?,
        })
    }
}

impl MailStore for IMAPStore {
    fn is_folder_complete(&self, folder: &str, state: &MailStoreState) -> bool {
        if let Some(f) = state.get_folder(folder) {
            let len = f.threads.len();
            if let Connection::ConnectionAndFolder(c) = &self.connection {
                return c.mailbox.exists as usize == len;
            }
        }
        false
    }
    fn list_folders(&mut self) -> Result<Vec<Name>> {
        let (folders, connection) = list_imap_folders(self.only_connection()?)?;
        self.connection = Connection::ConnectionOnly(connection);
        Ok(folders)
    }
    fn fetch_headers(&mut self, folder_path: &str, threads: &mut MessageThreader) -> Result<()> {
        let start = threads.len();
        let step = 40;
        let mut connection = self.examine(folder_path)?;
        let amount = step.min((connection.mailbox.exists as usize).saturating_sub(start));
        connection = fetch_headers(connection, start, amount, threads)?;
        self.connection = Connection::ConnectionAndFolder(connection);
        Ok(())
    }
    fn fetch(&mut self, folder: &str, uid: Uid) -> Result<Message> {
        let connection = self.examine(folder)?;
        let (message, connection) = fetch(connection, uid)?;
        self.connection = Connection::ConnectionAndFolder(connection);
        Ok(message)
    }
}

pub trait Stream: Read + Write {}
impl Stream for TcpStream {}
impl Stream for TlsStream<TcpStream> {}

pub fn connect_tcp(domain: &str, port: u16) -> Result<Box<Stream>> {
    let socket_addr = (domain, port);
    Ok(Box::new(TcpStream::connect(socket_addr)?))
}

pub fn connect_tls(domain: &str, port: u16) -> Result<Box<Stream>> {
    let socket_addr = (domain, port);
    let connector = TlsConnector::builder().build()?;
    let stream = TcpStream::connect(socket_addr)?;
    Ok(Box::new(connector.connect(domain, stream)?))
}

pub fn connect(config: &IMAPConfig) -> Result<Client<Box<Stream>>> {
    let stream = if config.ssl {
        connect_tls(&config.domain, config.port)?
    } else {
        connect_tcp(&config.domain, config.port)?
    };
    let mut imap_socket = Client::new(stream);
    imap_socket.login(&config.username, &config.password)?;
    Ok(imap_socket)
}

pub fn list_imap_folders(mut connection: ConnectionOnly) -> Result<(Vec<Name>, ConnectionOnly)> {
    let mut folders = Vec::new();

    match connection.0.list("", "*") {
        Ok(f) => for f in &f {
            folders.push(Name {
                delimiter: f.delimiter().into(),
                name: f.name().into(),
            });
        },
        Err(e) => println!("Error listing: {}", e),
    }
    folders.sort_by(|a, b| a.name.cmp(&b.name));
    Ok((folders, connection))
}

fn parse_message(atts: &[AttributeValue]) -> Result<Message> {
    if atts.len() != 2 {
        return Err(format!("Atts length is not 2 but {}", atts.len()).into());
    }
    let uid = if let AttributeValue::Uid(uid) = atts[0] {
        uid as usize
    } else {
        return Err(format!("First item in atts is not a Uid but {:?}.", atts[0]).into());
    };
    let header_data = if let AttributeValue::BodySection {
        section: Some(SectionPath::Full(MessageSection::Header)),
        index: None,
        data: Some(data),
    } = &atts[1]
    {
        data
    } else {
        return Err(format!("Second item in atts is not a headers but {:?}.", atts[1]).into());
    };
    Ok(Message::from_headers(Uid(uid), true, header_data)?)
}

pub fn fetch_headers(
    mut connection: ConnectionAndFolder,
    start: usize,
    amount: usize,
    threads: &mut MessageThreader,
) -> Result<ConnectionAndFolder> {
    // alternative:
    // FETCH 1:* BODY.PEEK[HEADER.FIELDS (DATE FROM SUBJECT MESSAGE-ID REFERENCES)]
    let response_bytes = connection.client.run_command_and_read_response(&format!(
        "FETCH {}:{} (UID BODY.PEEK[HEADER])",
        start + 1,
        start + amount
    ))?;
    let mut remainder: &[u8] = &response_bytes;
    loop {
        match parse_response(remainder) {
            IResult::Done(r, Response::Fetch(_, atts)) => {
                remainder = r;
                let message = Arc::new(parse_message(&atts)?);
                threads.add_message(message);
            }
            IResult::Done(_, r) => {
                return Err(format!("Unexpected result for parse_response {:?}", r).into());
            }
            IResult::Incomplete(needed) => {
                return Err(format!(
                    "fetch_header result is incomplete, needed {:?} more bytes to parse '{}'",
                    needed,
                    String::from_utf8_lossy(&response_bytes)
                ).into());
            }
            IResult::Error(e) => return Err(e.into()),
        };
        if remainder.is_empty() {
            return Ok(connection);
        }
    }
}

pub fn fetch(
    mut connection: ConnectionAndFolder,
    uid: Uid,
) -> Result<(Message, ConnectionAndFolder)> {
    let response_bytes = connection
        .client
        .run_command_and_read_response(&format!("UID FETCH {} (UID BODY.PEEK[])", uid))?;
    match parse_response(&response_bytes) {
        IResult::Done(_, Response::Fetch(_, atts)) => {
            if let AttributeValue::BodySection { data, .. } = &atts[1] {
                if let Some(data) = data {
                    return Ok((Message::new(uid, true, data.to_vec())?, connection));
                }
            }
            Err(format!("Unexpected result for parse_response {:?}", atts).into())
        }
        IResult::Done(_, r) => Err(format!("Unexpected result for parse_response {:?}", r).into()),
        IResult::Incomplete(needed) => Err(format!(
            "Fetch result is incomplete, needed {:?} more bytes to parse '{}'",
            needed,
            String::from_utf8_lossy(&response_bytes)
        ).into()),
        IResult::Error(e) => Err(e.into()),
    }
}
