use implementation::attachments::Attachments;
use interface::{AttachmentsEmitter, AttachmentsTrait, EmailEmitter, EmailTrait};
use mailparse::{parse_mail, ParsedMail};
use message::Message;
use std::sync::{Arc, Mutex};

pub struct EmailData {
    emit: EmailEmitter,
    attachments_emit: AttachmentsEmitter,
    pub email: Option<Arc<Message>>,
}

impl EmailData {
    pub fn set_email(&mut self, message: Option<Arc<Message>>) {
        if self.email != message {
            self.email = message;
            self.emit.subject_changed();
            self.emit.from_changed();
            self.emit.date_changed();
            self.emit.body_changed();
            self.emit.html_changed();
            self.attachments_emit.new_data_ready();
        }
    }
}

pub struct Email {
    emit: EmailEmitter,
    data: Arc<Mutex<EmailData>>,
    attachments: Attachments,
}

impl Email {
    pub fn data(&self) -> Arc<Mutex<EmailData>> {
        self.data.clone()
    }
    fn get_alternative(&self, mimetype: &str) -> String {
        let mut body = String::new();
        let data = self.data.lock().unwrap();
        if let Some(email) = &data.email {
            let email = parse_mail(&email.data).unwrap();
            let mut email = &email;
            //            print_structure(email);
            if let Some(alternative) = find_part("multipart/alternative", email) {
                email = alternative;
            }
            let alternative = find_part(mimetype, email);
            if let Some(alternative) = alternative {
                body = alternative.get_body().ok().unwrap_or_default();
            }
        }
        body
    }
}

impl EmailTrait for Email {
    fn new(emit: EmailEmitter, mut attachments: Attachments) -> Email {
        let data = Arc::new(Mutex::new(EmailData {
            emit: emit.clone(),
            email: None,
            attachments_emit: attachments.emit().clone(),
        }));
        attachments.set_email_data(Arc::clone(&data));
        Email {
            emit,
            data,
            attachments,
        }
    }
    fn emit(&self) -> &EmailEmitter {
        &self.emit
    }
    fn date<F>(&self, f: F)
    where
        F: FnOnce(&str),
    {
        if let Some(email) = &self.data.lock().unwrap().email {
            f(&email.date.to_rfc2822());
        } else {
            f("");
        }
    }
    fn from<F>(&self, f: F)
    where
        F: FnOnce(&str),
    {
        if let Some(email) = &self.data.lock().unwrap().email {
            f(&email.from);
        } else {
            f("");
        }
    }
    fn subject<F>(&self, f: F)
    where
        F: FnOnce(&str),
    {
        if let Some(email) = &self.data.lock().unwrap().email {
            f(&email.subject);
        } else {
            f("");
        }
    }
    fn body<F>(&self, f: F)
    where
        F: FnOnce(&str),
    {
        let plain = self.get_alternative("text/plain");
        f(&plain);
    }
    fn html<F>(&self, f: F)
    where
        F: FnOnce(&str),
    {
        let html = self.get_alternative("text/html");
        f(&html);
    }
    fn attachments(&self) -> &Attachments {
        &self.attachments
    }
    fn attachments_mut(&mut self) -> &mut Attachments {
        &mut self.attachments
    }
}

fn find_part<'a>(mimetype: &str, mail: &'a ParsedMail<'a>) -> Option<&'a ParsedMail<'a>> {
    if mail.ctype.mimetype == mimetype {
        return Some(mail);
    }
    mail.subparts.iter().find(
        |mail| mail.ctype.mimetype == mimetype,
    )
}
