use implementation::{Email, EmailData, FolderTreeListing, FolderTreeListingData, MailFolders,
                     MailFoldersData};
use interface::{MailModelEmitter, MailModelTrait};
use job_processor::{EmailReceiver, FoldersReceiver, Job, JobProcessor};
use mail_store::Name;
use message::{Message, Uid};
use std::path::PathBuf;
use std::sync::{Arc, Mutex};

pub struct MailModel {
    emit: MailModelEmitter,
    folders: MailFolders,
    folder_threads: FolderTreeListing,
    current_folder: String,
    job_processor: JobProcessor<R, F, E>,
    config_file: Option<PathBuf>,
    email: Email,
}

type R = Arc<Mutex<MailFoldersData>>;

impl FoldersReceiver for R {
    fn set_folders(&self, folders: Vec<Name>) {
        self.lock().unwrap().set_folders(folders)
    }
}

type F = Arc<Mutex<FolderTreeListingData>>;

type E = Arc<Mutex<EmailData>>;

impl EmailReceiver for E {
    fn set_email(&self, message: Option<Arc<Message>>) {
        self.lock().unwrap().set_email(message)
    }
}

impl MailModel {}

impl MailModelTrait for MailModel {
    fn new(
        emit: MailModelEmitter,
        email: Email,
        folder_threads: FolderTreeListing,
        folders: MailFolders,
    ) -> MailModel {
        MailModel {
            emit: emit.clone(),
            folders,
            folder_threads,
            current_folder: "/".into(),
            job_processor: JobProcessor::new(),
            email,
            config_file: None,
        }
    }
    fn emit(&self) -> &MailModelEmitter {
        &self.emit
    }
    fn folders(&self) -> &MailFolders {
        &self.folders
    }
    fn folders_mut(&mut self) -> &mut MailFolders {
        &mut self.folders
    }
    fn folder_threads(&self) -> &FolderTreeListing {
        &self.folder_threads
    }
    fn folder_threads_mut(&mut self) -> &mut FolderTreeListing {
        &mut self.folder_threads
    }
    fn current_folder(&self) -> &str {
        &self.current_folder
    }
    fn set_current_folder(&mut self, folder: String) {
        if self.current_folder == folder {
            return;
        }
        self.current_folder = folder;
        self.emit.current_folder_changed();
        if !self.current_folder.is_empty() {
            let _ = self.job_processor.send(Job::SetFolder(
                self.folder_threads.data(),
                self.current_folder.clone(),
            ));
        }
    }
    fn email(&self) -> &Email {
        &self.email
    }
    fn email_mut(&mut self) -> &mut Email {
        &mut self.email
    }
    fn set_email(&mut self, uid: u64) {
        let _ = self.job_processor.send(Job::SetEmail(
            self.email.data(),
            self.current_folder.clone(),
            Uid(uid as usize),
        ));
    }
    fn config_file(&self) -> Option<&str> {
        self.config_file.as_ref().and_then(|s| s.to_str())
    }
    fn set_config_file(&mut self, config_file: Option<String>) {
        self.config_file = config_file.clone().map(|s| s.into());
        if let Some(config_file) = &self.config_file {
            let _ = self.job_processor.send(Job::SetConfig(
                self.folders.data(),
                config_file.clone(),
            ));
        }
    }
}
